/*
  Copyright (C) 2001, 2002 Stephane Magnenat & Luc-Olivier de Charrière
    for any question or comment contact us at nct@ysagoon.com or nuage@ysagoon.com

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include <GUITextArea.h>
#include <assert.h>

TextArea::TextArea(int x, int y, int w, int h, const Font *font, bool readOnly, const char *text)
{
	this->readOnly=readOnly;
	this->x=x;
	this->y=y;
	this->w=w;
	this->h=h;
	this->font=font;
	textBuffer=NULL;
	assert(font);
	charHeight=font->getStringHeight(NULL);
	assert(charHeight);
	areaHeight=(h-8)/charHeight;
	areaPos=0;
	textBufferLength=0;
	
	cursorPos=0;
	cursorPosY=0;
	cursorScreenPosY=0;
	
	internalSetText(text);
}

TextArea::~TextArea(void)
{
	if (textBuffer)
		free(textBuffer);
}

void TextArea::internalPaint(void)
{
	assert(parent);
	assert(parent->getSurface());
	parent->getSurface()->setClipRect(x, y, w, h);
	parent->getSurface()->drawRect(x, y, w, h, 180, 180, 180);
	if (textBuffer)
	{
		for (unsigned i=0;(i<areaHeight)&&((signed)i<(signed)(lines.size()-areaPos));i++)
		{
			assert(i+areaPos<lines.size());
			if (i+areaPos<lines.size()-1)
			{
				unsigned tempLen=(unsigned)(lines[i+areaPos+1]-lines[i+areaPos]);
				char *temp=new char[tempLen+1];
				memcpy(temp, textBuffer+lines[i+areaPos], tempLen);
				temp[tempLen]=0;
				parent->getSurface()->drawString(x+4, y+4+(charHeight*i), w-8, font, "%s", temp);
				delete[] temp;
			}
			else
				parent->getSurface()->drawString(x+4, y+4+(charHeight*i), w-8, font, "%s", (textBuffer+lines[i+areaPos]));
		}
	}
	if (!readOnly)
	{
		parent->getSurface()->drawVertLine(x+4+cursorScreenPosY, y+4+(charHeight*(cursorPosY-areaPos)), charHeight, 255, 255, 255);
	}
	parent->getSurface()->setClipRect();
}

void TextArea::paint(void)
{
	internalPaint();
}

void TextArea::repaint(void)
{
	parent->paint(x, y, w, h);
	internalPaint();
	parent->addUpdateRect(x, y, w, h);
}

void TextArea::onSDLEvent(SDL_Event *event)
{
	if (event->type==SDL_KEYDOWN)
	{
		SDLKey sym=event->key.keysym.sym;
		SDLMod mod=event->key.keysym.mod;
		
		switch (sym)
		{
			case SDLK_DELETE:
			if (!readOnly)
			{
				if (cursorPos<textBufferLength)
				{
					unsigned len=getNextUTF8Char(textBuffer, cursorPos);
					remText(cursorPos, len-cursorPos);
					parent->onAction(this, TEXT_MODIFIED, 0, 0);
				}
			}
			break;
			
			case SDLK_BACKSPACE:
			if (!readOnly)
			{
				if (cursorPos)
				{
					unsigned newPos=getPrevUTF8Char(textBuffer, cursorPos);
					unsigned len=cursorPos-newPos;
					cursorPos=newPos;
					remText(newPos, len);
					parent->onAction(this, TEXT_MODIFIED, 0, 0);
				}
			}
			break;
			
			case SDLK_HOME:
			{
				if (SDL_GetModState() & KMOD_CTRL)
				{
					cursorPos=0;
				}
				else
				{
					cursorPos=lines[cursorPosY];
				}
				computeAndRepaint();
				parent->onAction(this, TEXT_CURSOR_MOVED, 0, 0);
			}
			break;
			
			case SDLK_END:
			{
				if (SDL_GetModState() & KMOD_CTRL)
				{
					cursorPos=textBufferLength;
				}
				else
				{
					if (cursorPosY<lines.size()-1)
						cursorPos=lines[cursorPosY+1]-1;
					else
						cursorPos=textBufferLength;
				}
				computeAndRepaint();
				parent->onAction(this, TEXT_CURSOR_MOVED, 0, 0);
			}
			break;
			
			case SDLK_PAGEUP:
			{
				if (areaPos>0)
				{
					assert(lines.size()>areaHeight);
					
					// compute new areaPos
					areaPos-=MIN(areaPos, areaHeight);

					// if in edit mode, replace cursor
					if (!readOnly)
					{
						// TODO : UTF8 clean cursor displacement in text
						unsigned int cursorPosX=cursorPos-lines[cursorPosY];
						unsigned int newPosY=cursorPosY>areaPos+areaHeight-2 ? areaPos+areaHeight-2 : cursorPosY;
						if (newPosY!=cursorPosY)
						{
							unsigned int newLineLen=lines[cursorPosY]-lines[newPosY];

							if (cursorPosX<newLineLen)
							{
								cursorPos=lines[newPosY]+cursorPosX;
							}
							else
							{
								cursorPos=lines[newPosY]+newLineLen-1;
							}

							parent->onAction(this, TEXT_CURSOR_MOVED, 0, 0);
						}
					}

					computeAndRepaint();
				}
			}
			break;
			
			case SDLK_PAGEDOWN:
			{
				if (lines.size()>=areaHeight)
				{
					if (areaPos<lines.size()-areaHeight)
					{
						// compute new areaPos
						areaPos+=MIN(lines.size()-areaHeight-areaPos, areaHeight);
						
						// if in edit mode, replace cursor
						if (!readOnly)
						{
							// TODO : UTF8 clean cursor displacement in text
							unsigned int cursorPosX=cursorPos-lines[cursorPosY];
							unsigned int newPosY=cursorPosY<areaPos+1 ? areaPos+1 : cursorPosY;
							if (newPosY!=cursorPosY)
							{
								unsigned int newLineLen;
								if (newPosY==lines.size()-1)
								{
									newLineLen=textBufferLength-lines[newPosY];
								}
								else
								{
									newLineLen=lines[newPosY+1]-lines[newPosY]-1;
								}
								
								if (cursorPosX<newLineLen)
								{
									cursorPos=lines[newPosY]+cursorPosX;
								}
								else
								{
									cursorPos=lines[newPosY]+newLineLen-1;
								}
								
								parent->onAction(this, TEXT_CURSOR_MOVED, 0, 0);
							}
						}
						
						computeAndRepaint();
					}
				}
			}
			break;
			
			case SDLK_UP:
			{
				if ((!readOnly) && (cursorPosY>0))
				{
					// TODO : UTF8 clean cursor displacement in text
					unsigned int cursorPosX=cursorPos-lines[cursorPosY];
					unsigned int newLineLen=lines[cursorPosY]-lines[cursorPosY-1];
					
					if (cursorPosX<newLineLen)
					{
						cursorPos=lines[cursorPosY-1]+cursorPosX;
					}
					else
					{
						cursorPos=lines[cursorPosY]-1;
					}
					
					computeAndRepaint();
					parent->onAction(this, TEXT_CURSOR_MOVED, 0, 0);
				}
			}
			break;
				
			case SDLK_DOWN:
			{
				if ((!readOnly) && (cursorPosY+1<lines.size()))
				{
					// TODO : UTF8 clean cursor displacement in text
					unsigned int cursorPosX=cursorPos-lines[cursorPosY];
					unsigned int newLineLen;
					
					if (cursorPosY==lines.size()-2)
					{
						newLineLen=textBufferLength-lines[cursorPosY+1];
					}
					else
					{
						newLineLen=lines[cursorPosY+2]-lines[cursorPosY+1]-1;
					}
					
					if (cursorPosX < newLineLen)
					{
						cursorPos=lines[cursorPosY+1]+cursorPosX;
					}
					else
					{
						cursorPos=lines[cursorPosY+1]+newLineLen;
					}
					
					computeAndRepaint();
					parent->onAction(this, TEXT_CURSOR_MOVED, 0, 0);
				}
			}
			break;
			
			case SDLK_LEFT:
			if (!readOnly)
			{
				if (mod&KMOD_CTRL)
				{
					bool cont=true;
					while ((cursorPos>0) && cont)
					{
						cursorPos=getPrevUTF8Char(textBuffer, cursorPos);
						switch (textBuffer[cursorPos])
						{
							case '.':
							case ' ':
							case '\t':
							case ',':
							case '\'':
							cont=false;
							default:
							break;
						}
					}
					computeAndRepaint();
					parent->onAction(this, TEXT_CURSOR_MOVED, 0, 0);
				}
				else
				{
					if (cursorPos>0)
					{
						cursorPos=getPrevUTF8Char(textBuffer, cursorPos);
						computeAndRepaint();
						parent->onAction(this, TEXT_CURSOR_MOVED, 0, 0);
					}
				}
			}
			break;
		
			case SDLK_RIGHT:
			if (!readOnly)
			{
				if (mod&KMOD_CTRL)
				{
					bool cont=true;
					while ((cursorPos<textBufferLength) && cont)
					{
						cursorPos=getNextUTF8Char(textBuffer, cursorPos);
						switch (textBuffer[cursorPos])
						{
							case '.':
							case ' ':
							case '\t':
							case ',':
							case '\'':
							cont=false;
							default:
							break;
						}
					}
					computeAndRepaint();
					parent->onAction(this, TEXT_CURSOR_MOVED, 0, 0);
				}
				else
				{
					if (cursorPos<textBufferLength)
					{
						cursorPos=getNextUTF8Char(textBuffer, cursorPos);
						computeAndRepaint();
						parent->onAction(this, TEXT_CURSOR_MOVED, 0, 0);
					}
				}
			}
			break;
			
			case SDLK_ESCAPE:
			parent->onAction(this, TEXT_CANCELED, 0, 0);
			break;
			
			case SDLK_RETURN:
			if (!readOnly)
			{
				addChar('\n');
			}
			break;
		
			default:
			if (!readOnly)
			{
				Uint16 c=event->key.keysym.unicode;
				if (c)
				{
					char utf8text[4];
					UCS16toUTF8(c, utf8text);
					addText(utf8text);
					parent->onAction(this, TEXT_MODIFIED, 0, 0);
				}
			}
			break;
		}
	}
	else if (event->type==SDL_MOUSEBUTTONDOWN)
	{
		if (isPtInRect(event->button.x, event->button.y, x, y, w, h))
		{
			if (event->button.button == 4)
			{
				scrollUp();
			}
			else if (event->button.button == 5)
			{
				scrollDown();
			}
		}
	}
	/*else if (event->type==SDL_MOUSEBUTTONDOWN)
	{
		if ((event->button.x>size.x+size.w+theme->xScroll) &&
			(event->button.x<size.x+size.w+theme->xScroll+theme->scrollUp->getW()) &&
			(event->button.y>size.y+theme->y1Scroll) &&
			(event->button.y<size.y+theme->y1Scroll+theme->scrollUp->getH()))
		{
			if (areaPos>0)
			{
				areaPos--;
			}
			return true;
		}
		else if ((event->button.x>size.x+size.w+theme->xScroll) &&
				(event->button.x<size.x+size.w+theme->xScroll+theme->scrollUp->getW()) &&
				(event->button.y>size.y+size.h+theme->y2Scroll) &&
				(event->button.y<size.y+size.h+theme->y2Scroll+theme->scrollUp->getH()))
		{
			if (areaPos<lines.size()-areaHeight-1)
			{
				areaPos++;
			}
			return true;
		}
	}*/
	//return false;
}

void TextArea::setCursorPos(unsigned pos)
{
	if (pos>textBufferLength)
		pos=textBufferLength;
	cursorPos=pos;
	computeAndRepaint();
}

void TextArea::computeAndRepaint(void)
{
	// The only variable which is always valid is cursorPos,
	// so now we recompute cursorPosY from it.
	// But it is guarantied that cursorPosY < lines.size();
	
	assert(cursorPosY >= 0);
	assert(cursorPosY < lines.size());
	assert(cursorPos >= 0);
	assert(cursorPos <= textBufferLength);
	
	if (!readOnly)
	{
		// increment it (if needed)
		while ((cursorPosY < lines.size()-1)  && (lines[cursorPosY+1] <= cursorPos))
			cursorPosY++;

		// decrement it (if needed)
		while (cursorPos < lines[cursorPosY])
			cursorPosY--;

		// make sure the cursor the visible window follow the cursor
		if (cursorPosY>0)
		{
			while (cursorPosY-1<areaPos)
				areaPos--;
		}
		else
		{
			while (cursorPosY<areaPos)
				areaPos--;
		}
		if (cursorPosY<lines.size()-1)
		{
			while (cursorPosY+1>=areaPos+areaHeight)
				areaPos++;
		}
		else
		{
			while (cursorPosY>=areaPos+areaHeight)
				areaPos++;
		}

		// TODO : UTF8 clean cursor displacement in text should lead to the removal of this code !!
		// we need to assert cursorPos will point on the beginning of a valid UTF8 char
		unsigned utf8CleanCursorPos = lines[cursorPosY];
		while (utf8CleanCursorPos + getNextUTF8Char(textBuffer[utf8CleanCursorPos]) <= cursorPos)
		{
			utf8CleanCursorPos += getNextUTF8Char(textBuffer[utf8CleanCursorPos]);
		}
		cursorPos = utf8CleanCursorPos;
		
		// compute displayable cursor Pos
		char temp[1024];
		unsigned cursorPosX = cursorPos-lines[cursorPosY];
		assert(cursorPosX < 1024);
		memcpy(temp, textBuffer+lines[cursorPosY], cursorPosX);
		temp[cursorPosX]=0;
		cursorScreenPosY=font->getStringWidth(temp);
	}
	
	// repaint
	repaint();
}


void TextArea::internalSetText(const char *text)
{
	assert(text);
	if (text)
	{
		if (textBuffer)
		{
			free(textBuffer);
			lines.clear();
		}
		textBufferLength=strlen(text);
		textBuffer=(char *)malloc(textBufferLength+1);
		unsigned pos=0;

		char temp[1024];
		int temppos;
		int lastWhite=-1;
		lines.push_back(0);
		
		temppos=0;
		temp[temppos]=0;
		
		// TODO : add getStringWidth with a number of char paramater to get ride of temp
		while (pos<textBufferLength)
		{
			while ((font->getStringWidth(temp)<w-4-font->getStringWidth("W"))&&(pos<textBufferLength)&&(text[pos]!='\n'))
			{
				if (text[pos]==' ')
					lastWhite=pos;
				textBuffer[pos]=text[pos];
				temp[temppos]=text[pos];
				temppos++;
				pos++;
				assert(temppos < 1024);
				temp[temppos]=0;
			}
			if (pos<textBufferLength)
			{
				if (text[pos]=='\n')
				{
					textBuffer[pos]='\n';
					pos++;
					lines.push_back(pos);
					temppos=0;
					assert(temppos < 1024);
					temp[temppos]=0;
				}
				else // line overflow
				{
					if ((lastWhite!=-1) && (lastWhite>(int)lines[lines.size()-1]))
					{
						pos=lastWhite;
						textBuffer[pos]='\n';
						pos++;
					}
					lines.push_back(pos);
					temppos=0;
					assert(temppos < 1024);
					temp[temppos]=0;
				}
			}
			else
			{
				textBuffer[pos]=0;
			}
		}
		if (pos==textBufferLength)
			textBuffer[pos]=0;
	}
}

void TextArea::setText(const char *text)
{
	internalSetText(text);
	computeAndRepaint();
}

void TextArea::addText(const char *text)
{
	assert(text);
	assert(cursorPos <= textBufferLength);
	assert(cursorPos >= 0);
	
	if (text)
	{
		int ts=strlen(text);

		char *temp=(char *)malloc(textBufferLength+ts+1);

		if (readOnly)
		{
			memcpy(temp, textBuffer, textBufferLength);
			memcpy(temp+textBufferLength, text, ts);
		}
		else
		{
			memcpy(temp, textBuffer, cursorPos);
			memcpy(temp+cursorPos, text, ts);
			memcpy(temp+cursorPos+ts, textBuffer+cursorPos, textBufferLength-cursorPos);
			cursorPos+=ts;
		}
		
		temp[textBufferLength+ts]=0;
		
		setText(temp);
		
		free(temp);
	}
}

void TextArea::remText(unsigned pos, unsigned len)
{
	if (pos < textBufferLength)
	{
		// if we wanna delete past the end
		if (pos + len >= textBufferLength)
			len = textBufferLength - pos;
		
		unsigned newLen = textBufferLength-len;
		char *temp=(char *)malloc(newLen+1);
		
		memcpy(temp, textBuffer, pos);
		memcpy(temp+pos, textBuffer+pos+len, textBufferLength-len-pos);
		temp[newLen] = 0;
		
		// make sure the cursor isn't past the end now
		if (cursorPos > newLen)
			cursorPos = newLen;
		// because we can decrease line count, let's zero the Y offset
		cursorPosY = 0;
		
		setText(temp);
		
		free(temp);
	}
}

void TextArea::addChar(const char c)
{
	char text[2];
	text[0] = c;
	text[1] = 0;
	addText(text);
}

void TextArea::scrollDown(void)
{
	if (lines.size()>=areaHeight)
	{
		if (areaPos<lines.size()-areaHeight-1)
		{
			areaPos++;
			repaint();
		}
	}
}

void TextArea::scrollUp(void)
{
	if (lines.size()>=areaHeight)
	{
		if (areaPos>0)
		{
			areaPos--;
			repaint();
		}
	}
}

void TextArea::scrollToBottom(void)
{
	while ((signed)areaPos<(signed)lines.size()-(signed)areaHeight-1)
	{
		areaPos++;
	}
	repaint();
}
